-module(kayos_utils).
-export(
  [
    optimise_uvs/1,optimise_uvs/2,get_bounds/1,safe_name/1,
    find_executable/1,split_list/2,cr_to_nl/1,db/1,db/2,
    foreach_except_last/3, foreach_except_last_incr/4,
    comma_separated_list/3,get_the_right_slashes/2,
    char_to_int/1
  ]).

-include("e3d.hrl").
-include_lib("kernel/include/file.hrl").

-define(KAYOS_DEBUG, false).

-import(lists, [append/2,reverse/1,seq/2,nth/2]).

db(String) ->
  case ?KAYOS_DEBUG of
    true -> io:put_chars(String ++ "\n");
    false -> ok end.
db(String,Var) ->
  case ?KAYOS_DEBUG of
    true -> io:format("~s: ~p~n",[String,Var]);
    false -> ok end.

char_to_int([Letter|T])->
  io:format("Letter ~s: ~p\n", [[Letter],Letter]),
  char_to_int(T);
char_to_int([]) -> ok.

get_the_right_slashes(String,OsGroup) ->
case OsGroup of
  win32 ->
  get_the_right_slashes(String,OsGroup,[]);
  _ -> String end.
get_the_right_slashes([Letter|String],OsGroup,NewString) ->
  NewLetter = case Letter of
    47 -> 92; % switch forward slash "/" for backslash "\"
    _ -> Letter end,
  get_the_right_slashes(String,OsGroup,[NewLetter|NewString]);
get_the_right_slashes([],_OsGroup,NewString) ->
  reverse(NewString).


safe_name(Name0) -> %this is currently a mess.
	case regexp:sub(Name0,"\s","_")of
		{ok,Name1,No} ->true;
		{error,Error} -> {error,Error}, Name1 = Name0;
		_ -> io:put_chars("Name: ~p\n", ["Oh crap"]), Name1 = Name0
	end,
	[N0|T] = Name1,
	N = case N0 >= "a" andalso N0 =< "z" of
		true -> N0 - "a" + "A";
		_ -> N0
	end,
	Name = append([N],T),
	Name.

get_bounds([{X,Y,Z}|T]) ->get_bounds(T, X, X, Y, Y, Z, Z).
get_bounds([{X,Y,Z}|T], Xmin0, Xmax0, Ymin0, Ymax0, Zmin0, Zmax0) ->
	case X of X when X < Xmin0 ->Xmin = X; _ -> Xmin = Xmin0 end,
	case X of X when X > Xmax0 ->Xmax = X; _ -> Xmax = Xmax0 end,
	case Y of Y when Y < Ymin0 ->Ymin = Y; _ -> Ymin = Ymin0 end,
	case Y of Y when Y > Ymax0 ->Ymax = Y; _ -> Ymax = Ymax0 end,
	case Z of Z when Z < Zmin0 ->Zmin = Z; _ -> Zmin = Zmin0 end,
	case Z of Z when Z > Zmax0 ->Zmax = Z; _ -> Zmax = Zmax0 end,
	get_bounds(T, Xmin, Xmax, Ymin, Ymax, Zmin, Zmax);
get_bounds([], Xmin, Xmax, Ymin, Ymax, Zmin, Zmax) ->
	{Xmin, Xmax, Ymin, Ymax, Zmin, Zmax}.

optimise_uvs(#e3d_mesh{fs=Fs,tx=Tx}=Mesh) -> optimise_uvs(Mesh,0).
optimise_uvs(#e3d_mesh{fs=Fs0,tx=Tx0}=Mesh, Tolerance) ->
	case weld_2d(Fs0,Tx0,Tolerance) of
	error -> error;
	{Fs,Tx} ->Mesh#e3d_mesh{fs=Fs,tx=Tx}
	end.

weld_2d(Fs,Tx0,Tol) ->
	NTxList = find_matches(Tx0,Tx0,Tx0,Tol),
	Tlist = seq(0, length(NTxList)-1),
	MissList = get_offsets(NTxList,Tlist),
	%io:format("\nNTxList: ~p\n", [NTxList]),
	%io:format("MissList: ~p\n", [MissList]),
	RemappedTx = remap_uvs(NTxList,Tlist,Tx0),
	FaceMap = make_facemap(NTxList, Tlist, list_to_tuple(NTxList), list_to_tuple(MissList)),
	%io:format("FaceMap: ~p\n", [FaceMap]),
	RemappedFs = remap_faces(Fs,list_to_tuple(FaceMap)),
	{RemappedFs, RemappedTx}.

get_offsets(L,L0) -> get_offsets(L,L0,[],0).
get_offsets([H|T],[H|T0], Acc0, Misses) ->
	Acc = append( Acc0, [Misses]),
	get_offsets( T, T0, Acc, Misses);
get_offsets([H|T],[H0|T0], Acc0, Misses) ->
	Acc = append(Acc0, [-1]), %pad list.
	get_offsets(T,T0,Acc,Misses+1);
get_offsets([],_,Acc,_) ->Acc;
get_offsets(_,[],Acc,_) ->Acc.



remap_faces(Fs,Fm) -> remap_faces(Fs,Fm,[],0).
remap_faces([#e3d_face{tx=Tx} = Fh|Ft], Fm, Acc, Incr) ->
	Tx0 = remap_face(Tx,Fm,Incr),
	remap_faces(Ft,Fm,[Fh#e3d_face{tx=Tx0}|Acc], Incr + length(Tx0));
remap_faces([],_,Acc,_) ->
	reverse(Acc).

remap_face(F,Fm,Incr) ->remap_face(F,Fm,[],Incr).
remap_face([Fh|Ft] = F, Fm,  Acc0, Incr)->
	E = element(Fh+1,Fm),
	Acc = append(Acc0, [E]),
	remap_face(Ft,Fm, Acc, Incr+1);
remap_face([],_,Acc,_) ->
	Acc.

make_facemap(L,L0,TxTpl,MTpl) -> make_facemap( L,L0,TxTpl,MTpl,[],0).
make_facemap([H|T], [H|T0], TxTpl, MTpl, Acc0, Matches) ->
	Acc = append(Acc0, [Matches]),
	make_facemap(T, T0, TxTpl, MTpl, Acc, Matches+1);
make_facemap([H|T],[H0|T0], TxTpl, MTpl, Acc0, Matches) ->
	E = element(H+1, TxTpl),
	E0 = element(E+1, MTpl),
	Acc = append(Acc0, [E-E0]),
	make_facemap(T,T0,TxTpl,MTpl, Acc,Matches);
make_facemap([],_,_,_,Acc,_) ->Acc;
make_facemap(_,[],_,_,Acc,_) ->Acc.

remap_uvs(N,N0,L) ->remap_uvs(N,N0,L,[],0).
remap_uvs([H|T],[H|T0],[Txh|Txt],Acc0,Incr) ->
	Acc = append(Acc0,[Txh]),
	remap_uvs(T,T0,Txt,Acc,Incr+1);
remap_uvs([H|T],[H0|T0],[Txh|Txt],Acc,Incr)->
	remap_uvs(T,T0,Txt,Acc,Incr);
remap_uvs([],_,_,Acc,_) ->Acc;
remap_uvs(_,[],_,Acc,_) ->Acc;
remap_uvs(_,_,[],Acc,_) ->Acc.

find_matches(Tx,Tx0,Txl,Tol) ->find_matches(Tx,Tx0,Txl,Tol,[],0).
find_matches([{U,V}=H|T]=Tx, [{U0,V0} = H0|T0] = Tx0,Txl,Tol,Acc0,Incr) ->
	case matches(U,U0,V,V0,Tol) of
		true ->
			Acc = append(Acc0, [Incr]),
			find_matches(T,Txl,Txl,Tol,Acc,0);
		false ->
			find_matches( Tx,T0,Txl,Tol,Acc0,Incr+1)
	end;
find_matches([],T,Txl,Tol,Acc,Incr) ->
	Acc;
find_matches(T,[],Txl,Tol,Acc,Incr) ->
	Acc.

matches(U,U0,V,V0,Tol) ->
	case U of
		U when U < (U0 - Tol) ->false;
		U when U =< (U0 +Tol) ->matches_v(V,V0,Tol);
		_ -> false
	end.

matches_v(V,V0,Tol) ->
	case V of
		V when V < (V0 - Tol) ->false;
		V when V =< (V0 + Tol) ->true;
		_-> false
	end.



%% Like os:find_executable, but if the found executable is a .bat file
%% on windows; scan the .bat file for a real executable file.
%%
find_executable(Name) ->
    case os:find_executable(Name) of
	false ->
	    false;
	Filename ->
	    case os:type() of
		{win32,_} ->
		    case lowercase(filename:extension(Filename)) of
			".bat" ->
			    find_in_bat(Filename);
			_ ->
			    Filename
		    end;
		_ ->
		    Filename
	    end
    end.

%% Search .bat file for an executable file
find_in_bat(Filename) ->
    case file:open(Filename, [read]) of
	{ok,F} ->
	    R = scan_bat(F),
	    file:close(F),
	    R;
	_ ->
	    false
    end.

%% Scan each line of the .bat file
scan_bat(F) ->
    case rskip_warg(skip_walpha(io:get_line(F, ""))) of
	"" ->
	    scan_bat(F);
	"echo"++[C|_] when C==$ ; C==$\t; C==$\n ->
	    scan_bat(F);
	"set"++[C|_] when C==$ ; C==$\t; C==$\n ->
	    scan_bat(F);
	Line when list(Line) ->
	    %% Check if this is the name of an executable file
	    Filename = filename:nativename(Line),
	    case file:read_file_info(Filename) of
		{ok,#file_info{type=regular,access=A,mode=M}} ->
		    case A of
			read when (M band 8#500) == 8#500 ->
			    Filename;
			read_write when (M band 8#500) == 8#500 ->
			    Filename;
			_ ->
			    scan_bat(F)
		    end;
		_ ->
		    scan_bat(F)
	    end;
	_ ->
	    false
    end.



%% Skip whitespace and one '@' from beginning of line

skip_walpha([$ |T]) ->
    skip_walpha(T);
skip_walpha([$\t|T]) ->
    skip_walpha(T);
skip_walpha([$\n|T]) ->
    skip_walpha(T);
skip_walpha([$@|T]) ->
    skip_walpha1(T);
skip_walpha(L) ->
    L.

skip_walpha1([$ |T]) ->
    skip_walpha(T);
skip_walpha1([$\t|T]) ->
    skip_walpha(T);
skip_walpha1([$\n|T]) ->
    skip_walpha(T);
skip_walpha1(L) ->
    L.

%% Skip whitespace and '%d' .bat file arguments from end of line

rskip_warg(L) ->
    rskip_warg1(lists:reverse(L)).

rskip_warg1([$ |T]) ->
    rskip_warg1(T);
rskip_warg1([$\t|T]) ->
    rskip_warg1(T);
rskip_warg1([$\n|T]) ->
    rskip_warg1(T);
rskip_warg1([D,$%|T]) when D >= $0, D =< $9 ->
    rskip_warg1(T);
rskip_warg1(L) ->
    lists:reverse(L).

%% Convert all A-Z in string to lowercase

lowercase([]) ->
    [];
lowercase([C|T]) when C >= $A, C =< $Z ->
    [(C + $a - $A)|lowercase(T)];
lowercase([C|T]) ->
    [C|lowercase(T)].

%% Convert lonely CR to NL, and CRLF to NL

cr_to_nl([$\r,$\n|T]) ->
    [$\n|cr_to_nl(T)];
cr_to_nl([$\r|T]) ->
    [$\n|cr_to_nl(T)];
cr_to_nl([C|T]) ->
    [C|cr_to_nl(T)];
cr_to_nl([]) ->
    [].

%% Split a list into a list of length Pos, and the tail

split_list(List, Pos) when list(List), integer(Pos), Pos >= 0 ->
    case split_list1(List, Pos, []) of
	{_,_}=Result ->
	    Result;
	Error ->
	    erlang:fault(Error, [List, Pos])
    end.

split_list1(List, 0, Head) ->
    {lists:reverse(Head),List};
split_list1([], _Pos, _) ->
    error;
split_list1([H|T], Pos, Head) ->
    split_list1(T, Pos-1, [H|Head]).

foreach_except_last(F, F_each, [H,H1|T]) ->
    F(H),
    F_each(H),
    foreach_except_last(F, F_each, [H1|T]);
foreach_except_last(F, F_each, [H]) ->
    F(H),
    ok.

comma_separated_list(F,Fun,[H,H1|T]) ->
  Fun(H),
  io:put_chars(F, ","),
  comma_separated_list(F,Fun,[H1|T]);
comma_separated_list(F,Fun,[H]) ->
  Fun(H),
  ok.

foreach_except_last_incr(F, F_each, Incr, [H,H1|T] ) ->
	F(H, Incr),
	F_each(H),
	foreach_except_last_incr(F, F_each, Incr+1,  [H1|T]);
foreach_except_last_incr(F, F_each, Incr, [H]) ->
	F(H,Incr),
	ok.
