-module(pov_ui).

-import(lists, [foldl/3,map/2,foreach/2,reverse/1,seq/2,
              flat_length/1,append/1,append/2,keydelete/3]).
-import(kayos_utils, [split_list/2,db/1,db/2]).
-import(wpc_pov, [get_var/1,get_pref/2,set_pref/1]).
-include("e3d.hrl").
-include("e3d_image.hrl").
-include("pov.hrl").

-export([export_dialog/1,menu/2,pref_edit/1,dialog/2,set_additional_preferences/1]).

%--------- MENU STUFF ---------------------

menu({file,export}, Menu) ->
  maybe_append (export, Menu, menu_entry(export));
menu({file,export_selected}, Menu) ->
  maybe_append (export, Menu, menu_entry(export));
menu({file,render}, Menu) ->
  maybe_append(render, Menu, menu_entry(render));
menu({edit,plugin_preferences}, Menu) ->
  Menu++menu_entry(pref);
menu(_, Menu) ->
  Menu.

maybe_append(Condition, Menu, PluginMenu) ->
  db("Condition", Condition),
  db("Dialogs", get_var(dialogs)),
  Value = case Condition of
    export -> get_var(dialogs);
    edit -> get_var(dialogs);
    render -> get_var(renderer);
    _ -> false end,
  case Value of
    false -> Menu;
    _ -> Menu ++ PluginMenu
  end.

menu_entry(render) ->
  [{"Povray (image)",?TAG,[option]}];
menu_entry(export) ->
  [{"Povray (.pov)",?TAG,[option]}];
menu_entry(pref) ->
  [{"Povray",?TAG}].

%------------- DIALOG STUFF --------------

dialog({material_editor_setup,Name,Mat}, Dialog) ->
  maybe_append(edit, Dialog, material_dialog(Name, Mat));
dialog({material_editor_result,Name,Mat}, Res) ->
%erlang:display({?MODULE,?LINE,[Name,Mat,Res]}),
R = material_result(Name, Mat, Res),
%erlang:display({?MODULE,?LINE,[R]}),
R;
dialog({light_editor_setup,Name,Ps}, Dialog) ->
  maybe_append(edit, Dialog, light_dialog(Name, Ps));
dialog({light_editor_result,Name,Ps0}, Res) ->
  light_result(Name, Ps0, Res);
dialog(_X, Dialog) ->
  io:format("~p\n", [{_X,Dialog}]),
  Dialog.

%------------- LIGHTS --------------------

light_dialog(Name, Ps) ->
  OpenGL = proplists:get_value(opengl, Ps, []),
  Povray = proplists:get_value(?TAG, Ps, []),
  Type = proplists:get_value(type, OpenGL, []),
  NoShadow = proplists:get_value(noshadow, Povray, false),
  AtmoInteraction = proplists:get_value(atmo_interaction, Povray, true),
  AtmoAttenuation = proplists:get_value(atmo_att, Povray, false),
  Falloff = proplists:get_value(pov_falloff, Povray, 0),
  FadeDistance = proplists:get_value(fade_distance, Povray, false),
  FadeDistanceAmt = proplists:get_value(fade_distance_amt, Povray, 0.5),
  [
    {hframe,[
      {vframe,[
        {"No Shadow", NoShadow, [{key,noshadow}]},
        {"Atmospheric Interaction", AtmoInteraction, [{key,atmo_interaction}]},
        {" Atmospheric Attenuation", AtmoAttenuation, [{key,atmo_att}]},
        {hframe, [
          {label,"Falloff"},
          {menu, [
            {"none", 0},
            {"linear", 1},
            {"quadratic", 2}
          ], Falloff,[{key,pov_falloff}]}
        ]},
        {hframe, [
          {"Fade Distance",FadeDistance,[{key,fade_distance}]},
          {text, FadeDistanceAmt, [{range, {0.0,100.0}},{key,fade_distance_amt}]}]}
      ],
      [{title,"Povray Options"},{minimized,true}]}|
      area_light_dialog(Name,Type,Ps)
    ]}|spot_light_dialog(Name,Type,Ps)
  ].

spot_light_dialog(_Name, Type, Ps) ->
  case Type of
   spot ->
    Cylinder = proplists:get_value(cylinder, Ps, false),
    Tightness = proplists:get_value(tightness, Ps, 0.0),
    [
      {vframe, [
        {"Cylinder", Cylinder, [{key,cylinder}]},
        {hframe, [
          {label,"Tightness"},
          {slider,{text, Tightness, [{range,{0.0,100.0}},{key, tightness}]}}
        ]}
      ],[{title,"Povray Spotlight Options"},{minimized,true}]}
    ];
  _ -> []
  end.

area_light_dialog(_Name, Type, Ps) ->
  case Type of
    ambient -> [];
    _ ->
      AreaLight = proplists:get_value(area_light, Ps, false),
      AreaWidth = proplists:get_value(area_width, Ps, 2.0),
      AreaLightsX = proplists:get_value(area_lights_x, Ps, 2),
      AreaHeight = proplists:get_value(area_height, Ps, 2.0),
      AreaLightsY = proplists:get_value(area_lights_y, Ps, 2),
      AreaAdaptive = proplists:get_value(area_adaptive, Ps, 1),
      AreaJitter = proplists:get_value(area_jitter, Ps, false),
      AreaCircular = proplists:get_value(area_circular, Ps, false),
      AreaOrient = proplists:get_value(area_orient, Ps, false),
      [
        {vframe, [
          {"Area Light", AreaLight, [{key,area_light}]},
          {hframe, [
            {label, "  "},
            {label,"Width "},
            {text, AreaWidth, [{range,{0.0,100.0}},{key, area_width}]},
            {label,"Lights"},
            {text, AreaLightsX, [{range,{0,100}},{key, area_lights_x}]}
          ]},
          {hframe, [
            {label, "  "},
            {label,"Height"},
            {text, AreaHeight, [{range,{0.0,100.0}},{key, area_height}]},
            {label,"Lights"},
            {text, AreaLightsY, [{range,{0,100}},{key, area_lights_y}]}
          ]},
          {hframe, [
            {label, "  "},
            {label, "Adaptive"},
            {text, AreaAdaptive, [{range,{0,32}},{key, area_adaptive}]}
          ]},
          {hframe, [
            {label, "  "},
            {"Jitter",AreaJitter,[{key,area_jitter}]},
            {"Circular",AreaCircular,[{key,area_circular}]},
            {"Orient",AreaOrient,[{key,area_orient}]}
          ]}
        ],[{title,"Povray Area Light Options"},{minimized,true}]}
      ]
  end.

light_result(_Name, Ps0, Res0) ->
  {LightPs0,Res1} = light_result(Res0),
  {LightPs1,Res2} = light_result(Res1),
  {LightPs2,Res} = light_result(Res2),
  LightPs3 = append(LightPs0,LightPs1),
  LightPs = append(LightPs3,LightPs2),
  Ps = [{?TAG,LightPs}|keydelete(?TAG, 1, Ps0)],
% erlang:display({?MODULE,?LINE,[Ps,Res]}),
  {Ps,Res}.


light_result([_,{noshadow,_}|_]=Res) ->
  split_list(Res, 7);
light_result([_,{cylinder,_}|_]=Res) ->
  split_list(Res, 3);
light_result([_,{area_light,_}|_]=Res) ->
  split_list(Res, 10 );
light_result(Tail) ->
%    erlang:display({?MODULE,?LINE,Tail}),
    {[],Tail}.

%------------- MATERIALS ------------------

material_dialog(_Name, Mat) ->

  Pov = proplists:get_value(?TAG, Mat, []),

  Diffusion = proplists:get_value(diffusion, Pov, 0.6),
  Brilliance = proplists:get_value(brilliance, Pov, 1.0),
  PhongHL = proplists:get_value(phong_hl, Pov, false),
  PhongAmt = proplists:get_value(phongamt, Pov, 0.5),
  PhongSize = proplists:get_value(phongsize, Pov, 40.0),
  SpecHL = proplists:get_value(spec_hl, Pov, true),
  SpecAmt = proplists:get_value(specamt, Pov, 0.5),
  Roughness = proplists:get_value(roughness, Pov, 0.05),
  MetalicHL = proplists:get_value(metalic, Pov, false),
  MetalicAmt = proplists:get_value(metalicamt, Pov, 1.0),

  SpecReflection = proplists:get_value(reflection, Pov, false),
  FresnelRef = proplists:get_value(fresnel, Pov, false),
  ConserveEnergy = proplists:get_value(conserve_energy, Pov, false),
  Falloff = proplists:get_value(falloff, Pov, false),
  FalloffAmt = proplists:get_value(falloff_amt, Pov, 0.5),
  MinReflection = proplists:get_value(min_reflection, Pov, false),
  MinReflectionAmt = proplists:get_value(min_reflection_amt, Pov, 0.5),
  MetalicRef = proplists:get_value(metalic_ref, Pov, false),
  MetalicRefAmt = proplists:get_value(metalic_ref_amt, Pov, 1.0),
  Exponent = proplists:get_value(exponent, Pov, false),

  ExponentAmt = proplists:get_value(exponent_amt, Pov, 0.5),
  Iridescence = proplists:get_value(iridescence, Pov, false),
  IridescenceAmt = proplists:get_value(iridescence_amt, Pov, 0.12),
  IridThickness = proplists:get_value(irid_thickness, Pov, false),
  IridThicknessAmt = proplists:get_value(irid_thickness_amt, Pov, 0.1),
  IridTurb = proplists:get_value(irid_turb, Pov, false),
  IridTurbAmt = proplists:get_value(irid_turb_amt, Pov, 0.3),
  Hollow = proplists:get_value(hollow, Pov, false),
  IOR = proplists:get_value(pov_ior, Pov, 1.0),
  Caustics = proplists:get_value(pov_caustics, Pov, 0.0),

  DispersionAmt = proplists:get_value(pov_dispersion_amt, Pov, 1.0),
  DispersionSamples = proplists:get_value(pov_dispersion_samples, Pov, 7),
  FadeDistance = proplists:get_value(fade_distance, Pov, 0.0),
  FadePower = proplists:get_value(fade_power, Pov, 0),
  FadeColor = proplists:get_value(fade_color, Pov, {0.0,0.0,0.0}),
  PovInterior = proplists:get_value(pov_interior, Pov, false),
  InteriorMedia= proplists:get_value(interior_media, Pov, false),
  InteriorAbsorption = proplists:get_value(interior_absorption, Pov, {0.0,0.0,0.0}),
  InteriorEmission = proplists:get_value(interior_emission, Pov, {0.0,0.0,0.0}),
  InteriorScattering = proplists:get_value(interior_scattering, Pov, {0.0,0.0,0.0}),

  ScatterType = proplists:get_value(scatter_type, Pov, 1),
  EccentricityAmt = proplists:get_value(eccentricity_amt, Pov, 0.0),
  ExtinctionAmt = proplists:get_value(extinction_amt, Pov, 1.0),
  %PovBumpmap = proplists:get_value(pov_bump_map, Pov, "filename"),
  %UseBumpmap = proplists:get_value(pov_use_bump_map, Pov, false),
  MinimizedFinish = proplists:get_value({pov_finish,minimized}, Pov, true),
  MinimizedInterior = proplists:get_value({pov_interior,minimized}, Pov, true),
  [
    {vframe, [
      {label_column, [
        {"Diffusion", {slider,{text,Diffusion,[{range,{0.0,1.0}},{key,diffusion}]}}},
        {"Brilliance",{slider,{text,Brilliance,[{range,{0.0,100.0}},{key,brilliance}]}}}
      ]},
      separator,
      {hframe, [
        {"Phong highlights", PhongHL, [{key,phong_hl}]},
        {text,PhongAmt,[{range,{0.0,1.0}},{key,phongamt}]},
        {label, "Phong Size"},
        {text,PhongSize,[{range,{1.0,250.0}},{key,phongsize}]}
      ]},
      {hframe, [
        {"Specular highlights", SpecHL, [{key,spec_hl}]},
        {text,SpecAmt,[{range,{0.0,1.0}},{key,specamt}]},
        {label,"roughness"},
	{text,Roughness,[{range,{0.0005,1.0}},{key,roughness}]}
      ]},
      {hframe, [
        {"Metalic", MetalicHL, [{key,metalic}]},
        {slider,{text,MetalicAmt,[{range,{0.0,1.0}},{key,metalicamt}]}}
      ]},
      separator,
      {hframe, [
        {"Specular reflection", SpecReflection, [{key,reflection}]},
        {"fresnel", FresnelRef, [{key,fresnel}]},
        {"conserve energy", ConserveEnergy, [{key,conserve_energy}]}
      ]},
      {hframe, [
        {label, "  "},
        {vframe, [
          {"min reflection", MinReflection, [{key,min_reflection}]},
          {"falloff", Falloff, [{key,falloff}]},
          {"exponent", Exponent, [{key,exponent}]},
          {"metalic", MetalicRef, [{key,metalic_ref}]}
        ]},
        {vframe, [
          {slider,{text,MinReflectionAmt,[{range,{0.0,1.0}},{key,min_reflection_amt}]}},
          {slider,{text,MinReflectionAmt,[{range,{0.0,1.0}},{key,min_reflection_amt}]}},
          {slider,{text,ExponentAmt,[{range,{0.0,1.0}},{key,exponent_amt}]}},
          {slider,{text,MetalicRefAmt,[{range,{0.0,1.0}},{key,metalic_ref_amt}]}}
        ]}
      ]},
      separator,
      {hframe, [
        {vframe, [
          {"iridescence", Iridescence, [{key, iridescence}]},
          {hframe, [{label, "  "},{"thickness", IridThickness, [{key, irid_thickness}]}]},
          {hframe, [{label, "  "},{"turbulence", IridTurb, [{key, irid_turb}]}]}
        ]},
        {vframe, [
          {slider,{text,IridescenceAmt,[{range,{0.0,1.0}},{key, iridescence_amt}]}},
          {slider,{text,IridThicknessAmt,[{range,{0.0,1.0}},{key, irid_thickness_amt}]}},
          {slider,{text,IridTurbAmt,[{range,{0.0,1.0}},{key, irid_turb_amt}]}}
        ]}
      ]}
    ], [{title,"Povray Finish Options"},{minimized,MinimizedFinish},{key,{pov_finish,minimized}}]},

  {vframe, [
    {hframe, [
      {"interior",PovInterior,[{key,pov_interior}]},
      {menu, [ {"Solid",false}, {"Hollow",true}], Hollow,[{key,hollow}]}
    ]},
    {hframe, [
      {label,"  "},
      {vframe, [
        {label_column, [
          {"IOR",{slider, {text,IOR,[{range,{0.0,3.5}},{key,pov_ior}]}}},
          {"Caustics", {slider, {text,Caustics,[{range,{0.0,1.0}},{key,pov_caustics}]}}},
          {"Dispersion", {slider, {text,DispersionAmt, [{range,{1.0,1.1}},{key,pov_dispersion_amt}]}}}
        ]},
        {hframe, [
          {label,"  "},{label,"Samples"},{label,"  "},{label,"  "},
          {text,DispersionSamples, [{range,{2,150}},{key,pov_dispersion_samples}]}
        ]},
        {hframe, [
          {label,"Fade"},{text, FadeDistance, [{range,{0.0,99.9}},{key,fade_distance}]},
          {label,"Power"},{menu, [{"none", 0}, {"linear", 1}, {"quadratic", 2} ], FadePower,[{key,fade_power}]},
          {color, FadeColor, [{key,fade_color}]}
        ]}
      ]}
     ]},
     separator,
      {hframe, [
        {"media", InteriorMedia, [{key, interior_media}]}
      ]},
      {hframe, [
        {label, "  "},{label,"Absorption"},
        {color,InteriorAbsorption,[{key,interior_absorbtion}]},
        {label,"Emission"},
      {color,InteriorEmission,[{key,interior_emission}]},
      {label,"Scattering"},
      {color,InteriorScattering,[{key,interior_scattering}]}
    ]},
    {hframe, [
      {label, "  "}, {label,"Scattering Type"},
      {menu, [
          {"isotropic", 1},
          {"Mie haze", 2},
          {"Mie murky", 3},
          {"Rayleigh", 4},
          {"Henyey-Greenstein", 5}
        ],
	ScatterType, [{key,scatter_type}
      ]}
    ]},
    {hframe, [
      {label, "    "},
      {label,"extinction"},
      {text, ExtinctionAmt, [{range,{0.0,5.0}},{key, extinction_amt}]},
      {label,"eccentricity"},
      {text, EccentricityAmt, [{range,{-5.0,5.0}},{key, eccentricity_amt}]}
    ]}
    ], [{title,"Povray Interior Options"},{minimized,MinimizedInterior},{key,{pov_interior,minimized}}]}
  ].

material_result(_Name, Ps0, Res0) ->
  {MatPs,Res} = split_list(Res0, 45),
  Ps = [{?TAG,MatPs}|keydelete(?TAG, 1, Ps0)],
  {Ps,Res}.


%------------- PLUGIN PREFERENCES ------------

pref_edit(St) ->
  Dialogs = get_pref(dialogs, ?DEF_DIALOGS),
  PovVersion = get_pref(pov_version, povray),
  Split = get_pref(split,single),
  Tmpdir = get_pref(pov_tmpdir, "/tmp"),
  PovPath = case get_var(pov_path) of
    undefined -> "none";
    PPath -> PPath end,
  Dialog = [
    {vframe, [
      {menu, [
          {"Disabled Dialogs",disabled},
	  {"Automatic Dialogs", auto},
	  {"Enabled Dialogs", enabled}],
        Dialogs,[{key,dialogs}]},
       separator,
      {menu, [
          {"Megapov 0.7",megapov},
	  {"MLPov 0.8", mlpov},
	  {"POVRay for unix", povray},
	  {"POVRay for Windows", pvengine}],
	PovVersion, [{key,pov_version}]},
      {hframe,[{label,"  "},{"Use Subdivision Surface Suite", get_pref(use_sss, false), [{key,use_sss}]}]},
       separator,
      {menu, [
          {"create a single file", single},
	  {"create a file for each section", bysection},
	  {"create a file for each object", byobject}],
	Split,[{key,split}]},
    {"Export World", get_pref(world, true), [{key,world}]},
    {hframe, [
      {label, "  "},
      {"Use file", get_pref(predefined_world, false), [{key,predefined_world}]},
      {text, get_pref(world_file, "location of world definition"), [{key,world_file}]}
    ]},
    {"Export Camera", get_pref(camera, true), [{key,camera}]},
    {hframe, [
      {label, "  "},
      {"Use file", get_pref(predefined_camera, false), [{key,predefined_camera}]},
      {text, get_pref(camera_file, "location of camera definition"), [{key,camera_file}]}
    ]},
    {"Export Lights", get_pref(lights, true), [{key,lights}]},
    {hframe, [
      {label, "  "},
      {"Use file", get_pref(predefined_lights, false), [{key,predefined_lights}]},
      {text, get_pref(lights_file, "location of lights definition"), [{key,lights_file}]}
    ]},
    {"Export Materials", get_pref(materials, true), [{key,materials}]},
    {hframe, [
      {label, "  "},
      {"Use file", get_pref(predefined_materials, false), [{key,predefined_materials}]},
      {text, get_pref(materials_file, "location of materials"), [{key,materials_file}]}
    ]},
      separator,
      {hframe, [
        {label,"temporary files"},
	{text,Tmpdir,[{key,pov_tmpdir}]}
      ]},
      {label, "Location of POV-Ray program"},
      {hframe,[{label,"  "},{text,PovPath,[{key,pov_path}]}]}
    ]}
  ],
  wpa:dialog("Povray Options", Dialog, fun (Attr) -> pref_result(Attr,St) end).

pref_result(Attr, St) ->
  set_pref(Attr),
  wpc_pov:init_pref(),
  St.

%------------- EXPORT OPTIONS -------------
export_dialog(export) ->
common_dialog();

export_dialog(render) ->
 [ {vframe, [
    {hframe, [
      {label, "Width"},
      {text, get_pref(frame_width, 640), [{range, {0,100000}}, {key,frame_width}]},
      {label, "Height"},
      {text, get_pref(frame_height, 480), [{range, {0,100000}},{key,frame_height}]}
    ]}
  ]}|
  common_dialog()].

common_dialog() ->
  SubDiv = get_pref(subdivisions, 0),
  GI_Preset = get_pref(gi_preset, default),
  BGColor = get_pref(bg_color, {0.0,0.0,0.0}),

  [
    separator,
    {"Use Orthographic Camera", get_pref(ortho_cam, false), [{key,ortho_cam}]},
    separator,
    {"Export UV Coordinates", get_pref(export_uv, true), [{key,export_uv}]},
    {hframe, [
      {label,"   "},
      {"Optimise UVs", get_pref(optimise_uv, true), [{key,optimise_uvs}]}
    ]},
    {"Export Normals", get_pref(export_normals,true),[{key,export_normals}]},
    {hframe, [
      {label, "Sub-division Steps"},{slider, {text,SubDiv,[{range,{0,5}},{key,subdivisions}]}}
    ]},
    separator,
    {hframe, [
      {"Background Colour", get_pref(background,false), [{key,background}]},
      {color,BGColor, [{key,bg_color}]}
    ]},
    separator,
    {"Global Illumination", get_pref(radiosity, false), [{key,radiosity}]},
    {hframe, [
      {label, "  "},
      {menu, [
          {"Default", default},
	  {"Debug", debug},
	  {"Fast", fast},
	  {"Normal", normal},
	  {"Two Bounces", twobounce},
	  {"Final", final},
	  {"Outdoor Low Quality", outdoorlq},
	  {"Outdoor High Quality", outdoorhq},
	  {"Outdoor Light", outdoorlight},
	  {"Indoor Low Quality", indoorlq},
	  {"Indoor High Quality", indoorhq}
	  ],
        GI_Preset,[{key,gi_preset}]}
    ]}
  ].

set_additional_preferences(Attr) ->
  [
    {predifined_world,get_pref(predifined_world,true)},
    {world_file,get_pref(world_file,false)},
    {predifined_lights,get_pref(predifined_lights,true)},
    {lights_file,get_pref(lights_file,false)},
    {predifined_camera,get_pref(predifined_camera,true)},
    {camera_file,get_pref(camera_file,false)},
    {predifined_materials,get_pref(predifined_materials,true)},
    {materials_file,get_pref(materials_file,false)}|Attr
  ].
